# flake8: noqa
# Our flake extension misfires on type comments in strings below.

from lib2to3.tests.test_fixers import FixerTestCase


class AnnotateFromSignatureTestCase(FixerTestCase):

    def setTestData(self, data):
        raise NotImplementedError

    def test_basic(self):
        self.setTestData(
            [{"func_name": "nop",
              "path": "<string>",
              "line": 3,
              "signature": {
                  "arg_types": ["Foo", "Bar"],
                  "return_type": "Any"},
              }])
        a = """\
            class Foo: pass
            class Bar: pass
            def nop(foo, bar):
                return 42
            """
        b = """\
            from typing import Any
            class Foo: pass
            class Bar: pass
            def nop(foo, bar):
                # type: (Foo, Bar) -> Any
                return 42
            """
        self.check(a, b)

    def test_decorator_func(self):
        self.setTestData(
            [{"func_name": "foo",
              "path": "<string>",
              "line": 2,
              "signature": {
                  "arg_types": [],
                  "return_type": "int"},
              }])
        a = """\
            @dec
            def foo():
                return 42
            """
        b = """\
            @dec
            def foo():
                # type: () -> int
                return 42
            """
        self.check(a, b)

    def test_decorator_method(self):
        self.setTestData(
            [{"func_name": "Bar.foo",
              "path": "<string>",
              "line": 3,
              "signature": {
                  "arg_types": [],
                  "return_type": "int"},
              }])
        a = """\
            class Bar:
                @dec
                @dec2
                def foo(self):
                    return 42
            """
        b = """\
            class Bar:
                @dec
                @dec2
                def foo(self):
                    # type: () -> int
                    return 42
            """
        self.check(a, b)

    def test_nested_class_func(self):
        self.setTestData(
            [{"func_name": "A.B.foo",
              "path": "<string>",
              "line": 3,
              "signature": {
                  "arg_types": ['str'],
                  "return_type": "int"},
              }])
        a = """\
            class A:
                class B:
                    def foo(x):
                        return 42
            """
        b = """\
            class A:
                class B:
                    def foo(x):
                        # type: (str) -> int
                        return 42
            """
        self.check(a, b)

    def test_nested_func(self):
        self.setTestData(
            [{"func_name": "A.foo.bar",
              "path": "<string>",
              "line": 3,
              "signature": {
                  "arg_types": [],
                  "return_type": "int"},
              }])
        a = """\
            class A:
                def foo():
                    def bar():
                        return 42
            """
        b = """\
            class A:
                def foo():
                    def bar():
                        # type: () -> int
                        return 42
            """
        self.check(a, b)

    def test_keyword_only_argument(self):
        self.setTestData(
            [{"func_name": "nop",
              "path": "<string>",
              "line": 3,
              "signature": {
                  "arg_types": ["Foo", "Bar"],
                  "return_type": "Any"},
              }])
        a = """\
            class Foo: pass
            class Bar: pass
            def nop(foo, *, bar):
                return 42
            """
        b = """\
            from typing import Any
            class Foo: pass
            class Bar: pass
            def nop(foo, *, bar):
                # type: (Foo, Bar) -> Any
                return 42
            """
        self.check(a, b)

    def test_add_typing_import(self):
        self.setTestData(
            [{"func_name": "nop",
              "path": "<string>",
              "line": 1,
              # Check with and without 'typing.' prefix
              "signature": {
                  "arg_types": ["List[typing.AnyStr]", "Callable[[], int]"],
                  "return_type": "object"},
              }])
        a = """\
            def nop(foo, bar):
                return 42
            """
        b = """\
            from typing import AnyStr
            from typing import Callable
            from typing import List
            def nop(foo, bar):
                # type: (List[AnyStr], Callable[[], int]) -> object
                return 42
            """
        self.check(a, b)

    def test_typing_import_parens(self):
        self.setTestData(
            [{"func_name": "nop",
              "path": "<string>",
              "line": 1,
              # Check with and without 'typing.' prefix
              "signature": {
                  "arg_types": ["List[typing.AnyStr]", "Callable[[], int]"],
                  "return_type": "object"},
              }])
        a = """\
            from typing import (AnyStr, Callable, List)
            def nop(foo, bar):
                return 42
            """
        b = """\
            from typing import (AnyStr, Callable, List)
            def nop(foo, bar):
                # type: (List[AnyStr], Callable[[], int]) -> object
                return 42
            """
        self.check(a, b)

    def test_add_other_import(self):
        self.setTestData(
            [{"func_name": "nop",
              "path": "mod1.py",
              "line": 1,
              "signature": {
                  "arg_types": ["mod1.MyClass", "mod2.OtherClass"],
                  "return_type": "mod3.AnotherClass"},
              }])
        a = """\
            from mod3 import AnotherClass
            def nop(foo, bar):
                return AnotherClass()
            class MyClass: pass
            """
        b = """\
            from mod3 import AnotherClass
            from typing import TYPE_CHECKING
            if TYPE_CHECKING:
                from mod2 import OtherClass
            def nop(foo, bar):
                # type: (MyClass, OtherClass) -> AnotherClass
                return AnotherClass()
            class MyClass: pass
            """
        self.check(a, b)

    def test_add_other_import_safe(self):
        self.setTestData(
            [{"func_name": "nop",
              "path": "mod1.py",
              "line": 1,
              "signature": {
                  "arg_types": ["mod1.MyClass", "mod3.OtherClass"],
                  "return_type": "mod3.AnotherClass"},
              }])
        a = """\
            from mod3 import AnotherClass
            def nop(foo, bar):
                return AnotherClass()
            class MyClass: pass
            """
        b = """\
            from mod3 import AnotherClass
            from mod3 import OtherClass
            def nop(foo, bar):
                # type: (MyClass, OtherClass) -> AnotherClass
                return AnotherClass()
            class MyClass: pass
            """
        self.check(a, b)

    def test_type_by_import(self):
        self.setTestData(
            [{"func_name": "nop",
              "path": "mod1.py",
              "line": 1,
              "signature": {
                  "arg_types": ["mod1.MyClass"],
                  "return_type": "mod2.AnotherClass"},
              }])
        a = """\
            import mod2
            def nop(foo):
                return mod2.AnotherClass()
            class MyClass: pass
            """
        b = """\
            import mod2
            def nop(foo):
                # type: (MyClass) -> mod2.AnotherClass
                return mod2.AnotherClass()
            class MyClass: pass
            """
        self.check(a, b)

    def test_type_by_mod_import_as(self):
        self.setTestData(
            [{"func_name": "nop",
              "path": "mod1.py",
              "line": 1,
              "signature": {
                  "arg_types": ["mod1.MyClass"],
                  "return_type": "mod2.AnotherClass"},
              }])
        a = """\
            import mod2 as bar
            def nop(foo):
                return bar.AnotherClass()
            class MyClass: pass
            """
        b = """\
            import mod2 as bar
            def nop(foo):
                # type: (MyClass) -> bar.AnotherClass
                return bar.AnotherClass()
            class MyClass: pass
            """
        self.check(a, b)

    def test_type_by_dotted_import_as(self):
        self.setTestData(
            [{"func_name": "nop",
              "path": "mod1.py",
              "line": 1,
              "signature": {
                  "arg_types": ["mod1.MyClass"],
                  "return_type": "pkg.mod2.AnotherClass"},
              }])
        a = """\
            import pkg.mod2.AnotherClass as bar
            def nop(foo):
                return bar()
            class MyClass: pass
            """
        b = """\
            import pkg.mod2.AnotherClass as bar
            def nop(foo):
                # type: (MyClass) -> bar
                return bar()
            class MyClass: pass
            """
        self.check(a, b)

    def test_type_by_dotted_import_mod_as(self):
        self.setTestData(
            [{"func_name": "nop",
              "path": "mod1.py",
              "line": 1,
              "signature": {
                  "arg_types": ["mod1.MyClass"],
                  "return_type": "pkg.mod2.AnotherClass"},
              }])
        a = """\
            import pkg.mod2 as bar
            def nop(foo):
                return bar.AnotherClass()
            class MyClass: pass
            """
        b = """\
            import pkg.mod2 as bar
            def nop(foo):
                # type: (MyClass) -> bar.AnotherClass
                return bar.AnotherClass()
            class MyClass: pass
            """
        self.check(a, b)

    def test_type_by_import_as(self):
        self.setTestData(
            [{"func_name": "nop",
              "path": "mod1.py",
              "line": 1,
              "signature": {
                  "arg_types": ["mod1.MyClass"],
                  "return_type": "mod2.AnotherClass"},
              }])
        a = """\
            from mod2 import AnotherClass as bar
            def nop(foo):
                return bar()
            class MyClass: pass
            """
        b = """\
            from mod2 import AnotherClass as bar
            def nop(foo):
                # type: (MyClass) -> bar
                return bar()
            class MyClass: pass
            """
        self.check(a, b)

    def test_type_by_from_import_as_with_dotted_package(self):
        self.setTestData(
            [{"func_name": "nop",
              "path": "mod1.py",
              "line": 1,
              "signature": {
                  "arg_types": ["mod1.MyClass"],
                  "return_type": "pkg.mod2.AnotherClass"},
              }])
        a = """\
            from pkg.mod2 import AnotherClass as bar
            def nop(foo):
                return bar()
            class MyClass: pass
            """
        b = """\
            from pkg.mod2 import AnotherClass as bar
            def nop(foo):
                # type: (MyClass) -> bar
                return bar()
            class MyClass: pass
            """
        self.check(a, b)

    def test_parentheses_import(self):
        self.setTestData(
            [{"func_name": "nop",
              "path": "mod1.py",
              "line": 1,
              "signature": {
                  "arg_types": ["mod2.MyClass"],
                  "return_type": "mod2.AnotherClass"},
              }])
        a = """\
            from mod2 import (AnotherClass,
                              MyClass)
            def nop(foo):
                return AnotherClass()
            class MyClass: pass
            """
        b = """\
            from mod2 import (AnotherClass,
                              MyClass)
            def nop(foo):
                # type: (MyClass) -> AnotherClass
                return AnotherClass()
            class MyClass: pass
            """
        self.check(a, b)

    def test_type_checking_import(self):
        self.setTestData(
            [{"func_name": "nop",
              "path": "mod1.py",
              "line": 1,
              "signature": {
                  "arg_types": ["mod2.MyClass"],
                  "return_type": "mod3.AnotherClass"},
              }])
        a = """\
            from mod3 import AnotherClass
            from typing import TYPE_CHECKING
            if TYPE_CHECKING:
                import X
            def nop(foo):
                return AnotherClass()
            class MyClass: pass
            """
        b = """\
            from mod3 import AnotherClass
            from typing import TYPE_CHECKING
            if TYPE_CHECKING:
                import X
                from mod2 import MyClass
            def nop(foo):
                # type: (MyClass) -> AnotherClass
                return AnotherClass()
            class MyClass: pass
            """
        self.check(a, b)

    def test_type_checking_from_import(self):
        self.setTestData(
            [{"func_name": "nop",
              "path": "mod1.py",
              "line": 1,
              "signature": {
                  "arg_types": ["mod2.MyClass"],
                  "return_type": "mod3.AnotherClass"},
              }])
        a = """\
            from mod3 import AnotherClass
            from typing import TYPE_CHECKING
            if TYPE_CHECKING:
                import X
            def nop(foo):
                return AnotherClass()
            class MyClass: pass
            """
        b = """\
            from mod3 import AnotherClass
            from typing import TYPE_CHECKING
            if TYPE_CHECKING:
                import X
                from mod2 import MyClass
            def nop(foo):
                # type: (MyClass) -> AnotherClass
                return AnotherClass()
            class MyClass: pass
            """
        self.check(a, b)

    def test_add_kwds(self):
        self.setTestData(
            [{"func_name": "nop",
              "path": "<string>",
              "line": 1,
              "signature": {
                  "arg_types": ["int"],
                  "return_type": "object"},
              }])
        a = """\
            def nop(foo, **kwds):
                return 42
            """
        b = """\
            from typing import Any
            def nop(foo, **kwds):
                # type: (int, **Any) -> object
                return 42
            """
        self.check(a, b)

    def test_dont_add_kwds(self):
        self.setTestData(
            [{"func_name": "nop",
              "path": "<string>",
              "line": 1,
              "signature": {
                  "arg_types": ["int", "**AnyStr"],
                  "return_type": "object"},
              }])
        a = """\
            def nop(foo, **kwds):
                return 42
            """
        b = """\
            from typing import AnyStr
            def nop(foo, **kwds):
                # type: (int, **AnyStr) -> object
                return 42
            """
        self.check(a, b)

    def test_add_varargs(self):
        self.setTestData(
            [{"func_name": "nop",
              "path": "<string>",
              "line": 1,
              "signature": {
                  "arg_types": ["int"],
                  "return_type": "object"},
              }])
        a = """\
            def nop(foo, *args):
                return 42
            """
        b = """\
            from typing import Any
            def nop(foo, *args):
                # type: (int, *Any) -> object
                return 42
            """
        self.check(a, b)

    def test_dont_add_varargs(self):
        self.setTestData(
            [{"func_name": "nop",
              "path": "<string>",
              "line": 1,
              "signature": {
                  "arg_types": ["int", "*int"],
                  "return_type": "object"},
              }])
        a = """\
            def nop(foo, *args):
                return 42
            """
        b = """\
            def nop(foo, *args):
                # type: (int, *int) -> object
                return 42
            """
        self.check(a, b)

    def test_return_expr_not_none(self):
        self.setTestData(
            [{"func_name": "nop",
              "path": "<string>",
              "line": 1,
              "signature": {
                  "arg_types": [],
                  "return_type": "None"},
              }])
        a = """\
            def nop():
                return 0
            """
        b = """\
            from typing import Any
            from typing import Optional
            def nop():
                # type: () -> Optional[Any]
                return 0
            """
        self.check(a, b)

    def test_return_expr_none(self):
        self.setTestData(
            [{"func_name": "nop",
              "path": "<string>",
              "line": 1,
              "signature": {
                  "arg_types": [],
                  "return_type": "None"},
              }])
        a = """\
            def nop():
                return
            """
        b = """\
            def nop():
                # type: () -> None
                return
            """
        self.check(a, b)

    def test_generator_optional(self):
        self.setTestData(
            [{"func_name": "gen",
              "path": "<string>",
              "line": 1,
              "signature": {
                  "arg_types": [],
                  "return_type": "Optional[int]"},
              }])
        a = """\
            def gen():
                yield 42
            """
        b = """\
            from typing import Iterator
            def gen():
                # type: () -> Iterator[int]
                yield 42
            """
        self.check(a, b)

    def test_generator_plain(self):
        self.setTestData(
            [{"func_name": "gen",
              "path": "<string>",
              "line": 1,
              "signature": {
                  "arg_types": [],
                  "return_type": "int"},
              }])
        a = """\
            def gen():
                yield 42
            """
        b = """\
            from typing import Iterator
            def gen():
                # type: () -> Iterator[int]
                yield 42
            """
        self.check(a, b)

    def test_not_generator(self):
        self.setTestData(
            [{"func_name": "nop",
              "path": "<string>",
              "line": 1,
              "signature": {
                  "arg_types": [],
                  "return_type": "int"},
              }])
        a = """\
            def nop():
                def gen():
                    yield 42
            """
        b = """\
            def nop():
                # type: () -> int
                def gen():
                    yield 42
            """
        self.check(a, b)

    def test_add_self(self):
        self.setTestData(
            [{"func_name": "nop",
              "path": "<string>",
              "line": 1,
              "signature": {
                  "arg_types": [],
                  "return_type": "int"},
              }])
        a = """\
            def nop(self):
                pass
            """
        b = """\
            from typing import Any
            def nop(self):
                # type: (Any) -> int
                pass
            """
        self.check(a, b)

    def test_dont_add_self(self):
        self.setTestData(
            [{"func_name": "C.nop",
              "path": "<string>",
              "line": 1,
              "signature": {
                  "arg_types": [],
                  "return_type": "int"},
              }])
        a = """\
            class C:
                def nop(self):
                    pass
            """
        b = """\
            class C:
                def nop(self):
                    # type: () -> int
                    pass
            """
        self.check(a, b)

    def test_too_many_types(self):
        self.setTestData(
            [{"func_name": "nop",
              "path": "<string>",
              "line": 1,
              "signature": {
                  "arg_types": ["int"],
                  "return_type": "int"},
              }])
        a = """\
            def nop():
                pass
            """
        self.warns(a, a, "source has 0 args, annotation has 1 -- skipping", unchanged=True)

    def test_too_few_types(self):
        self.setTestData(
            [{"func_name": "nop",
              "path": "<string>",
              "line": 1,
              "signature": {
                  "arg_types": [],
                  "return_type": "int"},
              }])
        a = """\
            def nop(a):
                pass
            """
        self.warns(a, a, "source has 1 args, annotation has 0 -- skipping", unchanged=True)

    def test_classmethod(self):
        # Class methods need to work without a class name
        self.setTestData(
            [{"func_name": "nop",
              "path": "<string>",
              "line": 3,
              "signature": {
                  "arg_types": ["int"],
                  "return_type": "int"}
              }])
        a = """\
            class C:
                @classmethod
                def nop(cls, a):
                    return a
            """
        b = """\
            class C:
                @classmethod
                def nop(cls, a):
                    # type: (int) -> int
                    return a
            """
        self.check(a, b)

    def test_classmethod_named(self):
        # Class methods also should work *with* a class name
        self.setTestData(
            [{"func_name": "C.nop",
              "path": "<string>",
              "line": 3,
              "signature": {
                  "arg_types": ["int"],
                  "return_type": "int"}
              }])
        a = """\
            class C:
                @classmethod
                def nop(cls, a):
                    return a
            """
        b = """\
            class C:
                @classmethod
                def nop(cls, a):
                    # type: (int) -> int
                    return a
            """
        self.check(a, b)

    def test_staticmethod(self):
        # Static methods need to work without a class name
        self.setTestData(
            [{"func_name": "nop",
              "path": "<string>",
              "line": 3,
              "signature": {
                  "arg_types": ["int"],
                  "return_type": "int"}
              }])
        a = """\
            class C:
                @staticmethod
                def nop(a):
                    return a
            """
        b = """\
            class C:
                @staticmethod
                def nop(a):
                    # type: (int) -> int
                    return a
            """
        self.check(a, b)

    def test_staticmethod_named(self):
        # Static methods also should work *with* a class name
        self.setTestData(
            [{"func_name": "C.nop",
              "path": "<string>",
              "line": 3,
              "signature": {
                  "arg_types": ["int"],
                  "return_type": "int"}
              }])
        a = """\
            class C:
                @staticmethod
                def nop(a):
                    return a
            """
        b = """\
            class C:
                @staticmethod
                def nop(a):
                    # type: (int) -> int
                    return a
            """
        self.check(a, b)

    def test_long_form(self):
        self.maxDiff = None
        self.setTestData(
            [{"func_name": "nop",
              "path": "<string>",
              "line": 1,
              "signature": {
                  "arg_types": ["int", "int", "int",
                                "str", "str", "str",
                                "Optional[bool]", "Union[int, str]", "*Any"],
                  "return_type": "int"},
              }])
        a = """\
            def nop(a, b, c,  # some comment
                    d, e, f,  # multi-line
                              # comment
                    g=None, h=0, *args):
                return 0
            """
        b = """\
            from typing import Any
            from typing import Optional
            from typing import Union
            def nop(a,  # type: int
                    b,  # type: int
                    c,  # type: int  # some comment
                    d,  # type: str
                    e,  # type: str
                    f,  # type: str  # multi-line
                              # comment
                    g=None,  # type: Optional[bool]
                    h=0,  # type: Union[int, str]
                    *args  # type: Any
                    ):
                # type: (...) -> int
                return 0
            """
        self.check(a, b)

    def test_long_form_method(self):
        self.maxDiff = None
        self.setTestData(
            [{"func_name": "C.nop",
              "path": "<string>",
              "line": 2,
              "signature": {
                  "arg_types": ["int", "int", "int",
                                "str", "str", "str",
                                "Optional[bool]", "Union[int, str]", "*Any"],
                  "return_type": "int"},
              }])
        a = """\
            class C:
                def nop(self, a, b, c,  # some comment
                              d, e, f,  # multi-line
                                        # comment
                              g=None, h=0, *args):
                    return 0
            """
        b = """\
            from typing import Any
            from typing import Optional
            from typing import Union
            class C:
                def nop(self,
                        a,  # type: int
                        b,  # type: int
                        c,  # type: int  # some comment
                        d,  # type: str
                        e,  # type: str
                        f,  # type: str  # multi-line
                                        # comment
                        g=None,  # type: Optional[bool]
                        h=0,  # type: Union[int, str]
                        *args  # type: Any
                        ):
                    # type: (...) -> int
                    return 0
            """
        self.check(a, b)

    def test_long_form_classmethod(self):
        self.maxDiff = None
        self.setTestData(
            [{"func_name": "nop",
              "path": "<string>",
              "line": 3,
              "signature": {
                  "arg_types": ["int", "int", "int",
                                "str", "str", "str",
                                "Optional[bool]", "Union[int, str]", "*Any"],
                  "return_type": "int"},
              }])
        a = """\
            class C:
                @classmethod
                def nop(cls, a, b, c,  # some comment
                        d, e, f,
                        g=None, h=0, *args):
                    return 0
            """
        b = """\
            from typing import Any
            from typing import Optional
            from typing import Union
            class C:
                @classmethod
                def nop(cls,
                        a,  # type: int
                        b,  # type: int
                        c,  # type: int  # some comment
                        d,  # type: str
                        e,  # type: str
                        f,  # type: str
                        g=None,  # type: Optional[bool]
                        h=0,  # type: Union[int, str]
                        *args  # type: Any
                        ):
                    # type: (...) -> int
                    return 0
            """
        self.check(a, b)
        # Do the same test for staticmethod
        a = a.replace('classmethod', 'staticmethod')
        b = b.replace('classmethod', 'staticmethod')
        self.check(a, b)

    def test_long_form_trailing_comma(self):
        self.maxDiff = None
        self.setTestData(
            [{"func_name": "nop",
              "path": "<string>",
              "line": 3,
              "signature": {
                  "arg_types": ["int", "int", "int",
                                "str", "str", "str",
                                "Optional[bool]", "Union[int, str]"],
                  "return_type": "int"},
              }])
        a = """\
            def nop(a, b, c,  # some comment
                    d, e, f,
                    g=None, h=0):
                return 0
            """
        b = """\
            from typing import Optional
            from typing import Union
            def nop(a,  # type: int
                    b,  # type: int
                    c,  # type: int  # some comment
                    d,  # type: str
                    e,  # type: str
                    f,  # type: str
                    g=None,  # type: Optional[bool]
                    h=0,  # type: Union[int, str]
                    ):
                # type: (...) -> int
                return 0
            """
        self.check(a, b)

    def test_one_liner(self):
        self.setTestData(
            [{"func_name": "nop",
              "path": "<string>",
              "line": 1,
              "signature": {
                  "arg_types": ["int"],
                  "return_type": "int"},
              }])
        a = """\
            def nop(a):   return a
            """
        b = """\
            def nop(a):
                # type: (int) -> int
                return a
            """
        self.check(a, b)

    def test_variadic(self):
        self.setTestData(
            [{"func_name": "nop",
              "path": "<string>",
              "line": 1,
              "signature": {
                  "arg_types": ["Tuple[int, ...]"],
                  "return_type": "int"},
              }])
        a = """\
            def nop(a):   return 0
            """
        b = """\
            from typing import Tuple
            def nop(a):
                # type: (Tuple[int, ...]) -> int
                return 0
            """
        self.check(a, b)

    def test_nested(self):
        self.setTestData(
            [{"func_name": "nop",
              "path": "<string>",
              "line": 1,
              "signature": {
                  "arg_types": ["foo:A.B"],
                  "return_type": "None"},
              }])
        a = """\
            def nop(a):
                pass
            """
        b = """\
            from typing import TYPE_CHECKING
            if TYPE_CHECKING:
                from foo import A
            def nop(a):
                # type: (A.B) -> None
                pass
            """
        self.check(a, b)



# flake8: noqa
# Our flake extension misfires on type comments in strings below.

from lib2to3.tests.test_fixers import FixerTestCase

# deadcode: fix_annotate is used as part of the fixer_pkg for this test
from typeright.fixes import base


class TestFixAnnotate(FixerTestCase):

    def setUp(self):
        super(TestFixAnnotate, self).setUp(
            fix_list=["annotate_any"],
            fixer_pkg="typeright",
            options={
                'typeright': {
                    'annotation_style': 'py2',
                    'comment_style': 'auto',
                },
            },
        )

    def test_no_arg(self):
        a = """\
            def nop():
                return 42
            """
        b = """\
            from typing import Any
            def nop():
                # type: () -> Any
                return 42
            """
        self.check(a, b)

    def test_one_arg(self):
        a = """\
            def incr(arg):
                return arg+1
            """
        b = """\
            from typing import Any
            def incr(arg):
                # type: (Any) -> Any
                return arg+1
            """
        self.check(a, b)

    def test_two_args(self):
        a = """\
            def add(arg1, arg2):
                return arg1+arg2
            """
        b = """\
            from typing import Any
            def add(arg1, arg2):
                # type: (Any, Any) -> Any
                return arg1+arg2
            """
        self.check(a, b)

    def test_defaults(self):
        a = """\
            def foo(iarg=0, farg=0.0, sarg='', uarg=u'', barg=False):
                return 42
            """
        b = """\
            from typing import Any
            def foo(iarg=0, farg=0.0, sarg='', uarg=u'', barg=False):
                # type: (int, float, str, unicode, bool) -> Any
                return 42
            """
        self.check(a, b)

    def test_staticmethod(self):
        a = """\
            class C:
                @staticmethod
                def incr(self):
                    return 42
            """
        b = """\
            from typing import Any
            class C:
                @staticmethod
                def incr(self):
                    # type: (Any) -> Any
                    return 42
            """
        self.check(a, b)

    def test_classmethod(self):
        a = """\
            class C:
                @classmethod
                def incr(cls, arg):
                    return 42
            """
        b = """\
            from typing import Any
            class C:
                @classmethod
                def incr(cls, arg):
                    # type: (Any) -> Any
                    return 42
            """
        self.check(a, b)

    def test_instancemethod(self):
        a = """\
            class C:
                def incr(self, arg):
                    return 42
            """
        b = """\
            from typing import Any
            class C:
                def incr(self, arg):
                    # type: (Any) -> Any
                    return 42
            """
        self.check(a, b)

    def test_fake_self(self):
        a = """\
            def incr(self, arg):
                return 42
            """
        b = """\
            from typing import Any
            def incr(self, arg):
                # type: (Any, Any) -> Any
                return 42
            """
        self.check(a, b)

    def test_nested_fake_self(self):
        a = """\
            class C:
                def outer(self):
                    def inner(self, arg):
                        return 42
            """
        b = """\
            from typing import Any
            class C:
                def outer(self):
                    # type: () -> None
                    def inner(self, arg):
                        # type: (Any, Any) -> Any
                        return 42
            """
        self.check(a, b)

    def test_multiple_decorators(self):
        a = """\
            class C:
                @contextmanager
                @classmethod
                @wrapped('func')
                def incr(cls, arg):
                    return 42
            """
        b = """\
            from typing import Any
            class C:
                @contextmanager
                @classmethod
                @wrapped('func')
                def incr(cls, arg):
                    # type: (Any) -> Any
                    return 42
            """
        self.check(a, b)

    def test_stars(self):
        a = """\
            def stuff(*a, **kw):
                return 4, 2
            """
        b = """\
            from typing import Any
            def stuff(*a, **kw):
                # type: (*Any, **Any) -> Any
                return 4, 2
            """
        self.check(a, b)

    def test_idempotency(self):
        a = """\
            def incr(arg):
                # type: (Any) -> Any
                return arg+1
            """
        self.unchanged(a)

    def test_no_return_expr(self):
        a = """\
            def proc1(arg):
                return
            def proc2(arg):
                pass
            """
        b = """\
            from typing import Any
            def proc1(arg):
                # type: (Any) -> None
                return
            def proc2(arg):
                # type: (Any) -> None
                pass
            """
        self.check(a, b)

    def test_nested_return_expr(self):
        # The 'return expr' in inner() shouldn't affect the return type of outer().
        a = """\
            def outer(arg):
                def inner():
                    return 42
                return
            """
        b = """\
            from typing import Any
            def outer(arg):
                # type: (Any) -> None
                def inner():
                    # type: () -> Any
                    return 42
                return
            """
        self.check(a, b)

    def test_nested_class_return_expr(self):
        # The 'return expr' in class Inner shouldn't affect the return type of outer().
        a = """\
            def outer(arg):
                class Inner:
                    return 42
                return
            """
        b = """\
            from typing import Any
            def outer(arg):
                # type: (Any) -> None
                class Inner:
                    return 42
                return
            """
        self.check(a, b)

    def test_add_import(self):
        a = """\
            import typing
            from typing import Callable

            def incr(arg):
                return 42
            """
        b = """\
            import typing
            from typing import Callable
            from typing import Any

            def incr(arg):
                # type: (Any) -> Any
                return 42
            """
        self.check(a, b)

    def test_dont_add_import(self):
        a = """\
            def nop(arg=0):
                return
            """
        b = """\
            def nop(arg=0):
                # type: (int) -> None
                return
            """
        self.check(a, b)

    def test_long_form(self):
        self.maxDiff = None
        a = """\
            def nop(arg0, arg1, arg2, arg3, arg4,
                    arg5, arg6, arg7, arg8=0, arg9='',
                    *args, **kwds):
                return
            """
        b = """\
            from typing import Any
            def nop(arg0,  # type: Any
                    arg1,  # type: Any
                    arg2,  # type: Any
                    arg3,  # type: Any
                    arg4,  # type: Any
                    arg5,  # type: Any
                    arg6,  # type: Any
                    arg7,  # type: Any
                    arg8=0,  # type: int
                    arg9='',  # type: str
                    *args,  # type: Any
                    **kwds  # type: Any
                    ):
                # type: (...) -> None
                return
            """
        self.check(a, b)

    def test_long_form_trailing_comma(self):
        self.maxDiff = None
        a = """\
            def nop(arg0, arg1, arg2, arg3, arg4, arg5, arg6,
                    arg7=None, arg8=0, arg9='', arg10=False,):
                return
            """
        b = """\
            from typing import Any
            def nop(arg0,  # type: Any
                    arg1,  # type: Any
                    arg2,  # type: Any
                    arg3,  # type: Any
                    arg4,  # type: Any
                    arg5,  # type: Any
                    arg6,  # type: Any
                    arg7=None,  # type: Any
                    arg8=0,  # type: int
                    arg9='',  # type: str
                    arg10=False,  # type: bool
                    ):
                # type: (...) -> None
                return
            """
        self.check(a, b)

    def test_one_liner(self):
        a = """\
            class C:
                def nop(self, a): a = a; return a
                # Something
            # More
            pass
            """
        b = """\
            from typing import Any
            class C:
                def nop(self, a):
                    # type: (Any) -> Any
                    a = a; return a
                # Something
            # More
            pass
            """
        self.check(a, b)

    def test_idempotency_long_1arg(self):
        a = """\
            def nop(a  # type: int
                    ):
                pass
            """
        self.unchanged(a)

    def test_idempotency_long_1arg_comma(self):
        a = """\
            def nop(a,  # type: int
                    ):
                pass
            """
        self.unchanged(a)

    def test_idempotency_long_2args_first(self):
        a = """\
            def nop(a,  # type: int
                    b):
                pass
            """
        self.unchanged(a)

    def test_idempotency_long_2args_last(self):
        a = """\
            def nop(a,
                    b  # type: int
                    ):
                pass
            """
        self.unchanged(a)

    def test_idempotency_long_varargs(self):
        a = """\
            def nop(*a  # type: int
                    ):
                pass
            """
        self.unchanged(a)

    def test_idempotency_long_kwargs(self):
        a = """\
            def nop(**a  # type: int
                    ):
                pass
            """
        self.unchanged(a)



import contextlib
import os
import tempfile
import textwrap
import unittest
from typing import Iterator

from typeright.annotations.infer import InferError
from typeright.annotations.main import (generate_annotations_json,
                                         generate_annotations_json_string)


class TestMain(unittest.TestCase):
    def test_generation(self):
        # type: () -> None
        data = """
        [
            {
                "path": "pkg/thing.py",
                "line": 422,
                "func_name": "my_function",
                "type_comments": [
                    "(List[int], str) -> None"
                ],
                "samples": 3
            }
        ]
        """
        with self.temporary_file() as target_path:
            with self.temporary_json_file(data) as source_path:
                generate_annotations_json(source_path, target_path)
            with open(target_path) as target:
                actual = target.read()

        actual = actual.replace(' \n', '\n')
        expected = textwrap.dedent("""\
            [
                {
                    "func_name": "my_function",
                    "line": 422,
                    "path": "pkg/thing.py",
                    "samples": 3,
                    "signature": {
                        "arg_types": [
                            "List[int]",
                            "str"
                        ],
                        "return_type": "None"
                    }
                }
            ]""")
        assert actual == expected

    def test_ambiguous_kind(self):
        # type: () -> None
        data = """
        [
            {
                "path": "pkg/thing.py",
                "line": 422,
                "func_name": "my_function",
                "type_comments": [
                    "(List[int], str) -> None",
                    "(List[int], *str) -> None"
                ],
                "samples": 3
            }
        ]
        """
        with self.assertRaises(InferError) as e:
            with self.temporary_json_file(data) as source_path:
                generate_annotations_json(source_path, '/dummy')
        assert str(e.exception) == textwrap.dedent("""\
            Ambiguous argument kinds:
            (List[int], str) -> None
            (List[int], *str) -> None""")

    def test_generate_to_memory(self):
        # type: () -> None
        data = """
        [
            {
                "path": "pkg/thing.py",
                "line": 422,
                "func_name": "my_function",
                "type_comments": [
                    "(List[int], str) -> None"
                ],
                "samples": 3
            }
        ]
        """
        with self.temporary_json_file(data) as source_path:
            output_data = generate_annotations_json_string(source_path)
        assert output_data == [
            {
                "path": "pkg/thing.py",
                "line": 422,
                "func_name": "my_function",
                "signature": {
                    "arg_types": [
                        "List[int]",
                        "str"
                    ],
                    "return_type": "None"
                },
                "samples": 3
            }
        ]

        with self.temporary_json_file(data) as source_path:
            output_data = generate_annotations_json_string(source_path, only_simple=True)
        assert output_data == []

    def test_generate_simple_signatures(self):
        # type: () -> None
        data = """
        [
            {
                "path": "pkg/thing.py",
                "line": 422,
                "func_name": "complex_function",
                "type_comments": [
                    "(List[int], str) -> None"
                ],
                "samples": 3
            },
            {
                "path": "pkg/thing.py",
                "line": 9000,
                "func_name": "simple_function",
                "type_comments": [
                    "(int, str) -> None"
                ],
                "samples": 3
            }
        ]
        """
        with self.temporary_json_file(data) as source_path:
            output_data = generate_annotations_json_string(source_path, only_simple=True)
        assert output_data == [
            {
                "path": "pkg/thing.py",
                "line": 9000,
                "func_name": "simple_function",
                "signature": {
                    "arg_types": [
                        "int",
                        "str"
                    ],
                    "return_type": "None"
                },
                "samples": 3
            }
        ]

    @contextlib.contextmanager
    def temporary_json_file(self, data):
        # type: (str) -> Iterator[str]
        source = None
        try:
            with tempfile.NamedTemporaryFile(mode='w', delete=False) as source:
                source.write(data)
            yield source.name
        finally:
            if source is not None:
                os.remove(source.name)

    @contextlib.contextmanager
    def temporary_file(self):
        # type: () -> Iterator[str]
        target = None
        try:
            with tempfile.NamedTemporaryFile(mode='w', delete=False) as target:
                pass
            yield target.name
        finally:
            if target is not None:
                os.remove(target.name)



from __future__ import absolute_import, print_function

import json
import re
import shlex
import subprocess
from lib2to3.pytree import Node
from typing import Any, Dict, List, Optional, Tuple

from .fix_annotate_json import BaseFixAnnotateFromSignature

REG = re.compile(r'`-?\d+')


def cleanup(s, node):
    if s == 'Tuple[]':
        # this gets generated for statements like `return ()`
        # FIXME: touch_import('typing', 'Any')
        return 'Tuple[Any, ...]'
    else:
        # fix 'T`1' -> 'T'
        return REG.sub('', s)


class FixAnnotateCommand(BaseFixAnnotateFromSignature):
    """Inserts annotations based on a command run in a subprocess for each
    location.  The command is expected to output a json string in the same
    format output by `dmypy suggest` and `pyannotate_tool --type-info`
    """

    def get_command(self, funcname, filename, lineno):
        # type: (str, str, int) -> List[str]
        command = self.type_options['command']
        return shlex.split(command.format(filename=filename, lineno=lineno,
                                          funcname=funcname))

    def get_types(self, node, results, funcname):
        # type: (Node, Dict[str, Any], str) -> Optional[Tuple[List[str], str]]
        cmd = self.get_command(funcname, self.filename, node.get_lineno())
        try:
            out = subprocess.check_output(cmd, stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as err:
            # dmypy suggest exits 2 anytime it can't generate a suggestion,
            # even for somewhat expected cases like when --no-any is enabled:
            if err.returncode != 2:
                self.log_message("Line %d: Failed calling %r: %s" %
                                 (node.get_lineno(), cmd,
                                  err.output.rstrip().encode()))
            return None
        except OSError as err:
            self.log_message("Line %d: Failed calling %r: %s" %
                             (node.get_lineno(), cmd, err))
            return None

        data = json.loads(out)
        signature = data[0]['signature']
        return [cleanup(arg, node) for arg in signature['arg_types']], \
            cleanup(signature['return_type'], node)



# flake8: noqa
# Our flake extension misfires on type comments in strings below.

import json
import os
import tempfile

try:
    from unittest.mock import patch
except ImportError:
    from mock import patch

from typeright.fixes.fix_annotate_json import (BaseFixAnnotateFromSignature,
                                                FixAnnotateJson)
from typeright.fixes.tests import base_py3


class TestFixAnnotateJson(base_py3.AnnotateFromSignatureTestCase):

    def setUp(self):
        super(TestFixAnnotateJson, self).setUp(
            fix_list=["annotate_json"],
            fixer_pkg="typeright",
            options={
                'typeright': {
                    'annotation_style': 'py3',
                    'top_dir': '',
                },
            },
        )

    def setTestData(self, data):
        self.refactor.options['typeright']['type_info'] = data
        self.filename = data[0]["path"]

    def test_line_number_drift(self):
        self.setTestData(
            [{"func_name": "nop",
              "path": "<string>",
              "line": 10,
              "signature": {
                  "arg_types": [],
                  "return_type": "int"},
              }])
        a = """\
            def nop(a):
                pass
            """
        self.warns(a, a, "signature from line 10 too far away -- skipping", unchanged=True)

    @patch('typeright.fixes.fix_annotate_json.BaseFixAnnotateFromSignature.set_filename')
    def test_set_filename(self, mocked_set_filename):
        self.filename = "/path/to/fileA.py"
        # trigger the fixer to run, with no expected changes
        self.unchanged("")
        mocked_set_filename.assert_called_with("/path/to/fileA.py")

        self.filename = "/path/to/fileB.py"
        # trigger the fixer to run, with no expected changes
        self.unchanged("")
        mocked_set_filename.assert_called_with("/path/to/fileB.py")



""" Test suite for the code in fixer_utils """

# Local imports
from lib2to3.fixer_util import Attr, Call, Comma, Name, is_import, syms
from lib2to3.pgen2 import token
from lib2to3.pytree import Leaf, Node
from lib2to3.tests.support import TestCase
from lib2to3.tests.test_util import parse
from typing import NamedTuple

from .. import fixer_utils


def _to_binding(node, type):
    if node.type == type:
        return node
    if node.type == token.NAME:
        return None
    for child in node.children:
        if child.type == type:
            return child
        res = _to_binding(child, type)
        if res is not None and res.type == type:
            return res


class Test_get_import_info(TestCase):
    def get_import_info(self, string):
        node = parse(string)
        for type in [syms.import_name, syms.import_from]:
            n = _to_binding(node, type)
            if n is not None:
                node = n
                break
        return fixer_utils.get_import_info(node)

    def test(self):
        passing_tests = (({'': [('b', 'b')]}, "import b"),
                         ({'': [('b', 'c')]}, "import b as c"),
                         ({"a": [('b', 'a.b')]}, "import a.b"),
                         ({"a.b": [('c', 'a.b.c')]}, "import a.b.c"),
                         ({"a": [('b', 'c')]}, "import a.b as c"),
                         ({"a.x": [('b', 'c')]}, "import a.x.b as c"),
                         ({"a": [('b', 'b')]}, "from a import b"),
                         ({"a": [('b', 'b'), ('c', 'c'), ('d', 'd')]}, "from a import b, c, d"),
                         ({"a": [('b', 'b'), ('c', 'c'), ('d', 'd')]}, "from a import (b, c, d)"),
                         ({"a": [('b', 'b'), ('c', 'c'), ('d', 'd')]}, "from a import (b,\nc,\nd)"),
                         ({"a": [('b', 'c')]}, "from a import b as c"),
                         ({"a.x": [('b', 'b')]}, "from a.x import b"),
                         ({"a.x": [('b', 'c')]}, "from a.x import b as c"),
                         ({"a": [('b', 'a.b')], "x": [('c', 'x.c')], "y": [('d', 'y.d')]},
                          "import a.b, x.c, y.d"),
                         ({"typing": [('Any', 'Any'), ('Callable', 'Callable'),
                                      ('Deque', 'Deque'), ('Generic', 'Generic'),
                                      ('Iterable', 'Iterable'), ('Iterator', 'Iterator'),
                                      ('List', 'List'), ('Optional', 'Optional'),
                                      ('Tuple', 'Tuple'), ('TypeVar', 'TypeVar'),
                                      ('Union', 'Union'), ('overload', 'overload')]},
                          """from typing import (Any, Callable, Deque, Generic, Iterable, Iterator, List,
                                                 Optional, Tuple, TypeVar, Union, overload)"""))

        for imports, stmt in passing_tests:
            print(stmt)
            n = self.get_import_info(stmt)
            self.assertTrue(n.imports)
            for key in imports.keys():
                self.assertTrue(n.imports.get(key))
                ref = set(imports.get(key))
                res = n.imports.get(key)
                self.assertTrue(ref == res)


class Test_find_import_info(TestCase):
    def find_import_info(self, package, name, string):
        node = parse(string)
        return fixer_utils.find_import_info(package, name, node)

    def test(self):
        package, name, binding, string = ('mod2', 'AnotherClass', 'bar',
                                          """
                                          # Attempt to disguise as mod2.AnotherClass - before real import
                                          import something as AnotherClass
                                          import AnotherClass
                                          import AnotherClass as other
                                          import mod3.AnotherClass
                                          import mod3.AnotherClass as thing
                                          import mod2.stuff as AnotherClass
                                          import mod3.mod2.AnotherClass
                                          import mod3.mod2.AnotherClass as ox
                                          import mod3.mod2.dingo as AnotherClass
                                          from something import geese as AnotherClass
                                          from AnotherClass import goober
                                          from AnotherClass import goober as gobble
                                          from mod3 import AnotherClass
                                          from mod3 import AnotherClass as bucket
                                          from mod2 import stuff as AnotherClass
                                          from mod3.mod2 import AnotherClass
                                          from mod3.mod2 import AnotherClass as trex
                                          from mod3.mod2 import dingo as AnotherClass

                                          # real import
                                          import mod2.AnotherClass as bar

                                          # Attempt to disguise as mod2.AnotherClass - after real import
                                          import something2 as AnotherClass
                                          import AnotherClass
                                          import AnotherClass as other2
                                          import mod4.AnotherClass
                                          import mod4.AnotherClass as thing2
                                          import mod2.stuff2 as AnotherClass
                                          import mod4.mod2.AnotherClass
                                          import mod4.mod2.AnotherClass as ox2
                                          import mod4.mod2.dingo2 as AnotherClass
                                          from something2 import geese2 as AnotherClass
                                          from AnotherClass import goober2
                                          from AnotherClass import goober2 as gobble2
                                          from mod4 import AnotherClass
                                          from mod4 import AnotherClass as bucket2
                                          from mod4 import stuff2 as AnotherClass
                                          from mod4.mod2 import AnotherClass
                                          from mod4.mod2 import AnotherClass as trex2
                                          from mod4.mod2 import dingo2 as AnotherClass""")

        res = self.find_import_info(package, name, string)
        self.assertTrue(res)
        self.assertTrue(res.package == package)
        self.assertTrue(res.entry == name)
        self.assertTrue(res.binding == binding)


class Test_create_type_checking_import(TestCase):
    def create_type_checking_import(self, package, name, string):
        node = parse(string)
        fixer_utils.create_type_checking_import(package, name, node)
        return node

    def test(self):
        string = """
        import a
        import b.a as c
        import c.d as e
        from X import Y

        import Z as W

        import typing.TYPE_CHECKING
        if typing.TYPE_CHECKING:
            from bar import R
            from bar import S
            from bar import T

        import mod as bar
        import foo
        from foo.mod2 import MyClass
        """

        ref = """
        import a
        import b.a as c
        import c.d as e
        from X import Y

        import Z as W

        import typing.TYPE_CHECKING
        if typing.TYPE_CHECKING:
            from bar import R
            from bar import S
            from bar import T
            from bar import TestImport

        import mod as bar
        import foo
        from foo.mod2 import MyClass
        """
        ref = parse(ref)

        res = self.create_type_checking_import("bar", "TestImport", string)
        self.assertTrue(res)
        self.assertTrue(res == ref)



from __future__ import absolute_import, print_function

from typing import Any, List

import docutils.nodes
from docutils import io  # type: ignore
from docutils.core import Publisher
from docutils.nodes import Element, Node  # type: ignore
from docutils.readers.standalone import Reader as _Reader
from docutils.utils import SystemMessage

from typeright.docs.parsers import Arg


def _clean_type(s):
    return s.strip().replace('\n', ' ')


class Reader(_Reader):
    doc_logger = None

    def pass_to_format_logger(self, msg):
        if msg['type'] == 'ERROR':
            log = self.doc_logger.error
        elif msg['type'] == 'WARNING':
            log = self.doc_logger.warning
        else:
            return
        log(Element.astext(msg), line=msg['line'])

    def new_document(self):
        document = _Reader.new_document(self)
        document.reporter.stream = False
        document.reporter.attach_observer(self.pass_to_format_logger)
        return document


def publish_doctree(source, logger):
    # type: (str, Any) -> Node
    pub = Publisher(reader=None, parser=None, writer=None,
                    settings=None,
                    source_class=io.StringInput,
                    destination_class=io.NullOutput)
    pub.reader = Reader(None, 'restructuredtext')
    pub.reader.doc_logger = logger
    pub.set_writer('null')
    pub.parser = pub.reader.parser
    pub.process_programmatic_settings(None, None, None)
    pub.set_source(source, None)
    pub.set_destination(None, None)
    output = pub.publish(enable_exit_status=False)
    return pub.document


class RestDocstring(object):

    def __init__(self, docstring, logger):
        self.docstring = docstring
        self.logger = logger

    def parse(self):
        params = []
        returns = None
        yields = None

        try:
            document = publish_doctree(self.docstring, self.logger)
        except SystemMessage:
            return params, returns, yields

        for field in document.traverse(condition=docutils.nodes.field):
            field_name = field.traverse(condition=docutils.nodes.field_name)[0]
            data = field_name.astext()
            # data is e.g.  'type foo'
            parts = data.strip().split()
            if len(parts) == 1 and parts[0] == 'returns':
                # converted google/numpy format with named results:
                # * **result1** (*str*) -- Description of first item
                # * **result2** (*bool*)
                # * **result3** (*int*) -- Description of third item
                # * *other stuff that is not return value.*
                items = field.traverse(condition=docutils.nodes.list_item)
                if items:
                    assert returns is None
                    item_types = []  # type: List[str]
                    for item in items:
                        text = item.astext()
                        item_types.append(Arg(text, field.line))
                        returns = item_types

            elif len(parts) == 1 or (len(parts) == 2 and parts[0] == 'type'):
                paras = field.traverse(condition=docutils.nodes.paragraph)
                if not paras:
                    continue
                paragraph = paras[0]
                arg = Arg(_clean_type(paragraph.astext()), paragraph.line)
                if len(parts) == 2:
                    # e.g. :type foo: xxxx
                    kind, name = parts
                    params.append((name, arg))
                elif len(parts) == 1 and parts[0] == 'rtype':
                    # e.g. :rtype: xxxx
                    assert returns is None
                    returns = [arg]
                elif len(parts) == 1 and parts[0] == 'Yields':
                    # e.g. converted from numpy/google format
                    assert yields is None
                    yields = [arg]

        return params, returns, yields



"""Some (nearly) end-to-end testing."""

import json
import os
import re
import shutil
import sys
import tempfile
import unittest

import pytest

# There seems to be no way to have this work and type-check without an
# explicit version check. :-(
if sys.version_info[0] == 2:
    from cStringIO import StringIO
else:
    from io import StringIO

from typing import List

from typeright.__main__ import main as dunder_main


class TestDunderMain(unittest.TestCase):
    def setUp(self):
        # type: () -> None
        self.tempdirname = tempfile.mkdtemp()
        self.tempfiles = []  # type: List[str]
        self.olddir = os.getcwd()
        os.chdir(self.tempdirname)

    def tearDown(self):
        # type: () -> None
        os.chdir(self.olddir)
        shutil.rmtree(self.tempdirname)

    def write_file(self, name, data):
        # type: (str, str) -> None
        self.tempfiles.append(name)
        with open(name, 'w') as f:
            f.write(data)

    def test_help(self):
        # type: () -> None
        self.main_test(["--help"], r"^usage:", r"^$", 0)

    def test_preview(self):
        # type: () -> None
        self.prototype_test(write=False)

    def test_final(self):
        # type: () -> None
        self.prototype_test(write=True)
        with open('gcd.py') as f:
            lines = [line.strip() for line in f.readlines()]
        assert '# type: (int, int) -> int' in lines

    @pytest.mark.skip
    def test_bad_encoding_message(self):
        # type: () -> None
        source_text = "# coding: unknownencoding\ndef f():\n  pass\n"
        self.write_file('gcd.py', source_text)
        self.write_file('type_info.json', '[]')
        encoding_message = "Can't parse gcd.py: unknown encoding: unknownencoding"
        self.main_test(['gcd.py'],
                       r'\A\Z',
                       r'\A' + re.escape(encoding_message),
                       0)

    def prototype_test(self, write):
        # type: (bool) -> None
        type_info = [
            {
                "path": "gcd.py",
                "line": 1,
                "func_name": "gcd",
                "type_comments": [
                    "(int, int) -> int"
                ],
                "samples": 2
            }
        ]
        source_text = """\
def gcd(a, b):
    while b:
        a, b = b, a%b
    return a
"""
        stdout_expected = """\
--- gcd.py	(original)
+++ gcd.py	(refactored)
@@ -1,4 +1,5 @@
 def gcd(a, b):
+    # type: (int, int) -> int
     while b:
         a, b = b, a%b
     return a
"""
        if not write:
            stderr_expected = """\
Refactored gcd.py
Files that need to be modified:
gcd.py
NOTE: this was a dry run; use -w to write files
"""
        else:
            stderr_expected = """\
Refactored gcd.py
Files that were modified:
gcd.py
"""
        self.write_file('type_info.json', json.dumps(type_info))
        self.write_file('gcd.py', source_text)
        args = ['gcd.py']
        if write:
            args.append('-w')
        args.extend(['--type-info', 'type_info.json', '--annotation-style=py2'])
        self.main_test(args,
                       re.escape(stdout_expected) + r'\Z',
                       re.escape(stderr_expected) + r'\Z',
                       0)

    def main_test(self, args, stdout_pattern, stderr_pattern, exit_code):
        # type: (List[str], str, str, int) -> None
        save_stdout = sys.stdout
        save_stderr = sys.stderr
        stdout = StringIO()
        stderr = StringIO()
        try:
            sys.stdout = stdout
            sys.stderr = stderr
            dunder_main(args)
            code = 0
        except SystemExit as err:
            code = err.code
        finally:
            sys.stdout = save_stdout
            sys.stderr = save_stderr
        stdout_value = stdout.getvalue()
        stderr_value = stderr.getvalue()
        assert re.match(stdout_pattern, stdout_value)
        match = re.match(stderr_pattern, stderr_value)
        ## if not match: print("\nNah")
        # else: print("\nYa!")
        # print(stderr_value)
        ## import pdb; pdb.set_trace()
        assert code == exit_code



# flake8: noqa
# Our flake extension misfires on type comments in strings below.

import unittest
from lib2to3.tests.test_fixers import FixerTestCase

# deadcode: fix_annotate is used as part of the fixer_pkg for this test
from typeright.fixes import base


class TestFixAnnotate3(FixerTestCase):

    def setUp(self):
        super(TestFixAnnotate3, self).setUp(
            fix_list=["annotate_any"],
            fixer_pkg="typeright",
            options={
                'typeright': {
                    'annotation_style': 'py3',
                },
            },
        )

    def test_no_arg_1(self):
        a = """\
            def nop():
                return 42
            """
        b = """\
            from typing import Any
            def nop() -> Any:
                return 42
            """
        self.check(a, b)

    def test_no_arg_2(self):
        a = """\
            def nop(): return 42
            """
        b = """\
            from typing import Any
            def nop() -> Any: return 42
            """
        self.check(a, b)

    def test_no_arg_3(self):
        a = """\
            def nop(
                    ):
                return 42
            """
        b = """\
            from typing import Any
            def nop(
                    ) -> Any:
                return 42
            """
        self.check(a, b)

    def test_no_arg_4(self):
        a = """\
            def nop(
                    )   \
                    :
                return 42
            """
        b = """\
            from typing import Any
            def nop(
                    ) -> Any   \
                    :
                return 42
            """
        self.check(a, b)

    def test_no_arg_5(self):
        a = """\
            def nop(    # blah
                    ):  # blah
                return 42 # blah
            """
        b = """\
            from typing import Any
            def nop(    # blah
                    ) -> Any:  # blah
                return 42 # blah
            """
        self.check(a, b)

    def test_no_arg_6(self):
        a = """\
            def nop(    # blah
                    )   \
                    :   # blah
                return 42 # blah
            """
        b = """\
            from typing import Any
            def nop(    # blah
                    ) -> Any   \
                    :   # blah
                return 42 # blah
            """
        self.check(a, b)

    def test_one_arg_1(self):
        a = """\
            def incr(arg):
                return arg+1
            """
        b = """\
            from typing import Any
            def incr(arg: Any) -> Any:
                return arg+1
            """
        self.check(a, b)

    def test_one_arg_2(self):
        a = """\
            def incr(arg=0):
                return arg+1
            """
        b = """\
            from typing import Any
            def incr(arg: int = 0) -> Any:
                return arg+1
            """
        self.check(a, b)

    def test_one_arg_3(self):
        a = """\
            def incr( arg=0 ):
                return arg+1
            """
        b = """\
            from typing import Any
            def incr( arg: int = 0 ) -> Any:
                return arg+1
            """
        self.check(a, b)

    def test_one_arg_4(self):
        a = """\
            def incr( arg = 0 ):
                return arg+1
            """
        b = """\
            from typing import Any
            def incr( arg: int = 0 ) -> Any:
                return arg+1
            """
        self.check(a, b)

    def test_two_args_1(self):
        a = """\
            def add(arg1, arg2):
                return arg1+arg2
            """
        b = """\
            from typing import Any
            def add(arg1: Any, arg2: Any) -> Any:
                return arg1+arg2
            """
        self.check(a, b)

    def test_two_args_2(self):
        a = """\
            def add(arg1=0, arg2=0.1):
                return arg1+arg2
            """
        b = """\
            from typing import Any
            def add(arg1: int = 0, arg2: float = 0.1) -> Any:
                return arg1+arg2
            """
        self.check(a, b)

    def test_two_args_3(self):
        a = """\
            def add(arg1, arg2=0.1):
                return arg1+arg2
            """
        b = """\
            from typing import Any
            def add(arg1: Any, arg2: float = 0.1) -> Any:
                return arg1+arg2
            """

    def test_two_args_4(self):
        a = """\
            def add(arg1, arg2 = 0.1):
                return arg1+arg2
            """
        b = """\
            from typing import Any
            def add(arg1: Any, arg2: float = 0.1) -> Any:
                return arg1+arg2
            """
        self.check(a, b)
        self.check(a, b)

    def test_defaults_1(self):
        a = """\
            def foo(iarg=0, farg=0.0, sarg='', uarg=u'', barg=False):
                return 42
            """
        b = """\
            from typing import Any
            def foo(iarg: int = 0, farg: float = 0.0, sarg: str = '', uarg: unicode = u'', barg: bool = False) -> Any:
                return 42
            """
        self.check(a, b)

    def test_defaults_2(self):
        a = """\
            def foo(iarg=0, farg=0.0, sarg='', uarg=u'', barg=False, targ=(1,2,3)):
                return 42
            """
        b = """\
            from typing import Any
            def foo(iarg: int = 0, farg: float = 0.0, sarg: str = '', uarg: unicode = u'', barg: bool = False, targ: Any = (1,2,3)) -> Any:
                return 42
            """
        self.check(a, b)

    def test_defaults_3(self):
        a = """\
            def foo(iarg=0, farg, sarg='', uarg, barg=False, targ=(1,2,3)):
                return 42
            """
        b = """\
            from typing import Any
            def foo(iarg: int = 0, farg: Any, sarg: str = '', uarg: Any, barg: bool = False, targ: Any = (1,2,3)) -> Any:
                return 42
            """
        self.check(a, b)

    def test_staticmethod(self):
        a = """\
            class C:
                @staticmethod
                def incr(self):
                    return 42
            """
        b = """\
            from typing import Any
            class C:
                @staticmethod
                def incr(self: Any) -> Any:
                    return 42
            """
        self.check(a, b)

    def test_classmethod(self):
        a = """\
            class C:
                @classmethod
                def incr(cls, arg):
                    return 42
            """
        b = """\
            from typing import Any
            class C:
                @classmethod
                def incr(cls, arg: Any) -> Any:
                    return 42
            """
        self.check(a, b)

    def test_instancemethod(self):
        a = """\
            class C:
                def incr(self, arg):
                    return 42
            """
        b = """\
            from typing import Any
            class C:
                def incr(self, arg: Any) -> Any:
                    return 42
            """
        self.check(a, b)

    def test_fake_self(self):
        a = """\
            def incr(self, arg):
                return 42
            """
        b = """\
            from typing import Any
            def incr(self: Any, arg: Any) -> Any:
                return 42
            """
        self.check(a, b)

    def test_nested_fake_self(self):
        a = """\
            class C:
                def outer(self):
                    def inner(self, arg):
                        return 42
            """
        b = """\
            from typing import Any
            class C:
                def outer(self) -> None:
                    def inner(self: Any, arg: Any) -> Any:
                        return 42
            """
        self.check(a, b)

    def test_multiple_decorators(self):
        a = """\
            class C:
                @contextmanager
                @classmethod
                @wrapped('func')
                def incr(cls, arg):
                    return 42
            """
        b = """\
            from typing import Any
            class C:
                @contextmanager
                @classmethod
                @wrapped('func')
                def incr(cls, arg: Any) -> Any:
                    return 42
            """
        self.check(a, b)

    def test_stars_1(self):
        a = """\
            def stuff(*a):
                return 4, 2
            """
        b = """\
            from typing import Any
            def stuff(*a: Any) -> Any:
                return 4, 2
            """
        self.check(a, b)

    def test_stars_2(self):
        a = """\
            def stuff(a, *b):
                return 4, 2
            """
        b = """\
            from typing import Any
            def stuff(a: Any, *b: Any) -> Any:
                return 4, 2
            """
        self.check(a, b)

    def test_keywords_1(self):
        a = """\
            def stuff(**kw):
                return 4, 2
            """
        b = """\
            from typing import Any
            def stuff(**kw: Any) -> Any:
                return 4, 2
            """
        self.check(a, b)

    def test_keywords_2(self):
        a = """\
            def stuff(a, **kw):
                return 4, 2
            """
        b = """\
            from typing import Any
            def stuff(a: Any, **kw: Any) -> Any:
                return 4, 2
            """
        self.check(a, b)

    def test_keywords_3(self):
        a = """\
            def stuff(a, *b, **kw):
                return 4, 2
            """
        b = """\
            from typing import Any
            def stuff(a: Any, *b: Any, **kw: Any) -> Any:
                return 4, 2
            """
        self.check(a, b)

    def test_keywords_4(self):
        a = """\
            def stuff(*b, **kw):
                return 4, 2
            """
        b = """\
            from typing import Any
            def stuff(*b: Any, **kw: Any) -> Any:
                return 4, 2
            """
        self.check(a, b)

    def test_no_return_expr(self):
        a = """\
            def proc1(arg):
                return
            def proc2(arg):
                pass
            """
        b = """\
            from typing import Any
            def proc1(arg: Any) -> None:
                return
            def proc2(arg: Any) -> None:
                pass
            """
        self.check(a, b)

    def test_nested_return_expr(self):
        # The 'return expr' in inner() shouldn't affect the return type of outer().
        a = """\
            def outer(arg):
                def inner():
                    return 42
                return
            """
        b = """\
            from typing import Any
            def outer(arg: Any) -> None:
                def inner() -> Any:
                    return 42
                return
            """
        self.check(a, b)

    def test_nested_class_return_expr(self):
        # The 'return expr' in class Inner shouldn't affect the return type of outer().
        a = """\
            def outer(arg):
                class Inner:
                    return 42
                return
            """
        b = """\
            from typing import Any
            def outer(arg: Any) -> None:
                class Inner:
                    return 42
                return
            """
        self.check(a, b)

    def test_add_import(self):
        a = """\
            import typing
            from typing import Callable

            def incr(arg):
                return 42
            """
        b = """\
            import typing
            from typing import Callable
            from typing import Any

            def incr(arg: Any) -> Any:
                return 42
            """
        self.check(a, b)

    def test_dont_add_import(self):
        a = """\
            def nop(arg=0):
                return
            """
        b = """\
            def nop(arg: int = 0) -> None:
                return
            """
        self.check(a, b)

    def test_long_form(self):
        self.maxDiff = None
        a = """\
            def nop(arg0, arg1, arg2, arg3, arg4,
                    arg5, arg6, arg7, arg8=0, arg9='',
                    *args, **kwds):
                return
            """
        b = """\
            from typing import Any
            def nop(arg0: Any, arg1: Any, arg2: Any, arg3: Any, arg4: Any,
                    arg5: Any, arg6: Any, arg7: Any, arg8: int = 0, arg9: str = '',
                    *args: Any, **kwds: Any) -> None:
                return
            """
        self.check(a, b)

    def test_long_form_trailing_comma(self):
        self.maxDiff = None
        a = """\
            def nop(arg0, arg1, arg2, arg3, arg4, arg5, arg6,
                    arg7=None, arg8=0, arg9='', arg10=False,):
                return
            """
        b = """\
            from typing import Any
            def nop(arg0: Any, arg1: Any, arg2: Any, arg3: Any, arg4: Any, arg5: Any, arg6: Any,
                    arg7: Any = None, arg8: int = 0, arg9: str = '', arg10: bool = False,) -> None:
                return
            """
        self.check(a, b)

    def test_one_liner(self):
        a = """\
            class C:
                def nop(self, a): a = a; return a
                # Something
            # More
            pass
            """
        b = """\
            from typing import Any
            class C:
                def nop(self, a: Any) -> Any: a = a; return a
                # Something
            # More
            pass
            """
        self.check(a, b)

    def test_idempotency_long_1arg(self):
        a = """\
            def nop(a: int
                    ):
                pass
            """
        self.unchanged(a)

    def test_idempotency_long_1arg_comma(self):
        a = """\
            def nop(a: int,
                    ):
                pass
            """
        self.unchanged(a)

    def test_idempotency_long_2args_first(self):
        a = """\
            def nop(a: int,
                    b):
                pass
            """
        self.unchanged(a)

    def test_idempotency_long_2args_last(self):
        a = """\
            def nop(a,
                    b: int
                    ):
                pass
            """
        self.unchanged(a)

    def test_idempotency_long_varargs(self):
        a = """\
            def nop(*a: int
                    ):
                pass
            """
        self.unchanged(a)

    def test_idempotency_long_kwargs(self):
        a = """\
            def nop(**a: int
                    ):
                pass
            """
        self.unchanged(a)

    def test_idempotency_arg0_ret_value(self):
        a = """\
            def nop() -> int:
                pass
            """
        self.unchanged(a)

    def test_idempotency_arg1_ret_value(self):
        a = """\
            def nop(a) -> int:
                pass
            """
        self.unchanged(a)

    def test_idempotency_arg1_default_1(self):
        a = """\
            def nop(a: int=0):
                pass
            """
        self.unchanged(a)

    def test_idempotency_arg1_default_2(self):
        a = """\
            def nop(a: List[int]=[]):
                pass
            """
        self.unchanged(a)

    def test_idempotency_arg1_default_3(self):
        a = """\
            def nop(a: List[int]=[1,2,3]):
                pass
            """
        self.unchanged(a)
